/******************************************************************************
 *
 * PROJECT: Carnegie Mellon Planetary Rover Project
 *          Task Control Architecture 
 * 
 * (c) Copyright 1991 Christopher Fedor and Reid Simmons.  All rights reserved.
 * 
 * MODULE: communications
 *
 * FILE: comModule.c
 *
 * ABSTRACT:
 * 
 * Module Communications
 *
 * REVISION HISTORY 
 *
 * $Log: comModule.c,v $
 * Revision 1.60  1996/01/23  00:06:24  rich
 * Fixed memory leak when a module connects and disconnects.  Also fixed a
 * problem with using the direct connection flag.  This was introduced when
 * we added contexts for keeping track of the central server.
 *
 * Revision 1.59  1996/01/05  16:31:05  rich
 * Added windows NT port.
 *
 * Revision 1.58  1995/12/21  19:17:35  reids
 * For safety's sake, copy strings when registering messages, handlers and
 *    when connecting.
 *
 * Revision 1.57  1995/12/17  20:21:13  rich
 * Have free routines set pointers to NULL.
 * Removed old makefiles.
 *
 * Revision 1.56  1995/12/15  01:23:06  rich
 * Moved Makefile to Makefile.generic to encourage people to use
 * GNUmakefile.
 * Fixed a memory leak when a module is closed and some other small fixes.
 *
 * Revision 1.55  1995/11/03  03:04:18  rich
 * Changed msgFind to keep if from going into an infinite loop if there is no
 * central connection.  This only happens when an exit procedure that does
 * not exit is registered.  msgFind can now return NULL, so I added some
 * checks for the return value to keep modules from seg-faulting.
 *
 * Revision 1.54  1995/10/29  18:26:29  rich
 * Initial creation of 8.3. Added changes made after 8.2 branch was
 * created. These mostly have to do with context switching.
 * 
 * Revision 1.53.2.3  1995/12/19  04:26:37  reids
 * For safety's sake, copy strings when registering messages, handlers and
 *   when connecting.
 *
 * Revision 1.53.2.2  1995/11/02  22:42:13  reids
 * Fixed a problem with incorrectly setting modData.hostName
 *
 * Revision 1.53.2.1  1995/10/27  15:16:40  rich
 * Fixed problems with connecting to multiple central servers.
 *
 * Revision 1.53  1995/10/25  22:48:00  rich
 * Fixed problems with context switching.  Now the context is a separate
 * data structure accessed from the module data structure, using the
 * currentContext field.  GET_C_GLOBAL is used instead of GET_M_GLOBAL for
 * the context dependent fields.
 *
 * Revision 1.52  1995/10/07  19:07:02  rich
 * Pre-alpha release of tca-8.2.
 * Added PROJECT_DIR. Added tcaWillListen.
 * Only transmit broadcast messages when there is a handler to receive them.
 * All system messages now start with "tca_".  Old messages are also supported.
 *
 * Revision 1.51  1995/08/14  21:31:19  rich
 * Got rid of the "sharedBuffers" flag on the dataMessages.  It was not the
 * right solution, and possibly caused a memory leak.
 * Limit pending for TCA_TAPPED_MSG_VAR to 1.
 *
 * Revision 1.50  1995/08/06  16:43:43  reids
 * A bug existed in that two demon monitors that sent the same ID number
 * would conflict (causing the wrong one to fire).  This has been fixed, and
 * in the process, one of the hash-key functions was made a bit more general.
 *
 * Revision 1.49  1995/08/05  17:16:16  reids
 * Several important bug fixes:
 *   a) Found a memory leak in the tms (when nodes are unasserted)
 *   b) Fixed a problem with direct connections that would cause TCA to crash
 *      when a command or goal message was sent from an inform or a query.
 *      As part of that fix, all command and goal messages that are sent from
 *      informs or queries are now added to the root node of the task tree.
 *
 * Revision 1.48  1995/07/19  14:25:52  rich
 * Added display and dump to the central interface.
 * Fixed problem with direct querries not returning to the correct module.
 * Added Argv versions of provides and requires.
 *
 * Revision 1.47  1995/07/12  04:54:24  rich
 * Release of 8.0.
 * Fixed problems with sending between machines of different endien.
 *
 * Revision 1.46  1995/07/10  16:16:57  rich
 * Interm save.
 *
 * Revision 1.45  1995/07/06  21:15:47  rich
 * Solaris and Linux changes.
 *
 * Revision 1.44  1995/06/14  03:21:26  rich
 * Added DBMALLOC_DIR.
 * More support for DOS.  Fixed some problems with direct connections.
 *
 * Revision 1.43  1995/06/05  23:58:46  rich
 * Improve support of detecting broken pipes.  Add support for OSF 2.
 * Add return types to the global variable routines.
 *
 * Revision 1.42  1995/04/21  03:53:12  rich
 * Added central commands to kill the task tree and close a module.
 * Added tcaGetContext and tcaSetContext to support connections to multiple
 * central servers.  tcaConnectModules can be called multiple times.
 * Fixed a bug in the resource limit pending.
 * Created seperate routines to print help and option messages.
 *
 * Revision 1.41  1995/04/19  14:27:50  rich
 * Fixed problems with lisp encode/decode functions.
 * Added types int32 and int16 for use where the size of the integer matters.
 *
 * Revision 1.40  1995/04/09  20:30:07  rich
 * Added /usr/local/include and /usr/local/lib to the paths for compiling
 * for sunOS machines. (Support for new vendor OS).
 * Create a tca directory in /tmp and put the socket in there so others can
 * delete dead sockets.  The /tmp directory has the sticky bit set so you
 * can't delete files even if you have write permission on the directory.
 * Fixes to libc.h to use the new declarations in the gcc header files and
 * avoid problems with dbmalloc.
 *
 * Revision 1.39  1995/04/08  02:06:19  rich
 * Added waitForReplyFrom to be able to block on replies from only one
 * source.  Useful when querying for the msg info information.  Added a
 * tcaQueryCentral that only accepts input from the central server.  Fixed
 * timing problems with direct connections.
 *
 * Revision 1.38  1995/04/07  05:02:55  rich
 * Fixed GNUmakefiles to find the release directory.
 * Cleaned up libc.h file for sgi and vxworks.  Moved all system includes
 * into libc.h
 * Got direct queries to work.
 * Fixed problem in allocating/initializing generic mats.
 * The direct flag (-c) now mostly works.  Connect message has been extended to
 * indicate when direct connections are the default.
 * Problem with failures on sunOS machines.
 * Fixed problem where tcaError would not print out its message if logging had
 * not been initialized.
 * Fixed another memory problem in modVar.c.
 * Fixed problems found in by sgi cc compiler.  Many type problems.
 *
 * Revision 1.37  1995/04/05  19:10:36  rich
 * Adding Reid's Changes.
 * Consolidating the matrix code.
 *
 * Revision 1.36  1995/04/04  19:41:50  rich
 * Added sgi support.
 * Split low level com routines out to be used in devUtils.
 * Improved some error messages.
 * Added central switch to default to direct connections.  Does not work yet.
 * Fixed the vectorization code.
 *
 * Revision 1.35  1995/03/30  15:42:32  rich
 * DBMALLOC works.  To use "gmake -k -w DBMALLOC=DBMALLOC install"
 * Added simple list of strings data structure that can be passed via tca
 * messages.
 * Use the string list to maintain a global variable of messages with taps.
 * Tapped messages are not sent via direct connections.
 * Implemented code to vectorize data to be sent so that it does not have
 * to be copied.  Currently, only flat, packed data structures are
 * vectored.  This can now be easily extended.
 * Changed Boolean -> BOOLEAN for consistency and to avoid conflicts with x11.
 * Fixed bug were central would try and free the "***New Module***" and
 * "*** Unkown Host***" strings when a module crashed on startup.
 * Fixed a bug reported by Jay Gowdy where the code to find the size of a
 * variable lenght array would access already freed data when called from
 * tcaFreeData.
 *
 * Revision 1.34  1995/03/28  01:14:22  rich
 * - Added ability to log data with direct connections.  Also fixed some
 * problems with global variables. It now uses broadcasts for watching variables.
 * - Added preliminary memory recovery routines to handle out of memory
 * conditions.  It currently purges items from resource queues.  Needs to
 * be tested.
 * - If the CENTRALHOST environment variable is not set, try the current
 * host.
 * - Fixed a problem with central registered messages that caused the parsed
 * formatters to be lost.
 * - Added const declarations where needed to the prototypes in tca.h.
 * - tcaGetConnections: Get the fd_set.  Needed for direct connections.
 * - Added tcaExecute and tcaExecuteWithConstraints.  Can "execute" a goal
 *   or command.
 * - tcaPreloadMessage: Preload the definition of a message from the
 *   central server.
 *
 * Revision 1.33  1995/03/19  19:39:22  rich
 * Implemented direct connections using tcaDirectResouce call.
 * Also made the basics.h file a module include.
 * Changed class in the interval structure to be interval_class to avoid a
 * conflict with C++.
 *
 * Revision 1.32  1995/03/18  15:11:00  rich
 * Fixed updateVersion script so it can be run from any directory.
 *
 * Revision 1.31  1995/02/06  14:46:16  reids
 * Removed the "Global" suffix from tcaRootNode, tcaServer and tcaDefaultTime
 *
 * Revision 1.30  1995/01/30  16:17:40  rich
 * Added indications of message byte order and alignment to the message
 * header in the upper two bytes of the classID.
 * Now the receiver translates the byte order but must receive data in
 * "PACKED" alignment.
 * Made -gstabs come after -g for i386_mach machines so the debugger will work.
 *
 * Revision 1.29  1995/01/25  00:00:59  rich
 * Release of tca 7.9.  Mostly speed improvements.
 * The cvs binaries may now be located in /usr/local.
 * Fixed problems with little endian translation.
 *
 * Revision 1.28  1995/01/18  22:39:52  rich
 * TCA 7.9: Speed improvements.
 * Use unix sockets for communication on the same machine.
 * Eliminate copying.
 * Optimize loop for arrays, especially simple, primitive arrays.
 * Optimize the buffer size.
 *
 * Revision 1.27  1994/11/02  21:34:00  rich
 * Now works for linux machines (i486).
 * Got afs to work on alpha (and hopefully other vendor OS's)
 * Added generic Makefile.
 * Made libc.h and tcaMatrix.h module includes.
 * Reduced the size of libc.h by using more system includes.
 *
 * Revision 1.26  1994/11/02  16:14:26  reids
 * Fixed connectSocket to handle server hosts specified as <host>:<port>
 *
 * Revision 1.25  1994/10/25  17:09:41  reids
 * Changed the logging functions to accept variable number of arguments.
 *
 * Revision 1.24  1994/06/07  02:30:22  rich
 * Include the top level README and other files in the tarfile.
 * Include extra header infomation to get rid of warnings.
 *
 * Revision 1.23  1994/05/31  03:23:44  rich
 * Removed CFLAGS_sun4.
 * Removed cfree and bzero from mem routines.
 * Set zero wait on both sides of the pipe.  Can connect to host using inet
 * number.
 *
 * Revision 1.22  1994/05/25  04:57:06  rich
 * Defined macros for registering simple messages and handlers at once.
 * Added function to ignore logging for all messages associated with a
 * global variable.
 * Moved module global variable routines to a new file so they are not
 * included in the .sa library file.  Gets better code sharing and lets you
 * debug these routines.
 * Added code to force the module variables to be re-initialized after the
 * server goes down.
 * tcaClose now will not crash if the server is down and frees some module
 * memory.
 * The command line flag "-u" turns off the simple user interface.
 * Added routines to free hash tables and id tables.
 *
 * Revision 1.21  1994/05/24  13:48:31  reids
 * Fixed so that messages are not sent until a tcaWaitUntilReady is received
 * (and the expected number of modules have all connected)
 *
 * Revision 1.20  1994/05/17  23:15:20  rich
 * Added global variables and associated routines.
 * Added some error checking.  The central connection is now set to -1
 * rather than zero to prevent tca messages from being send to stdout.
 * Now compiles on the sgi machines.  Still need to have the endian and
 * alignment figured out automatically.
 *
 * Revision 1.19  1994/05/11  01:57:12  rich
 * Now set an invalid tcaServerGlobal (a socket fd) to -1 rather than 0
 * which is stdout.
 * Added checks to make sure tcaServerGlobal is a valid socket before
 * sending messages or waiting for messages.
 *
 * Revision 1.18  1994/05/05  00:46:03  rich
 * Added a gmake makefile GNUmakefile so that the system can be easily
 * compiled on different machines.
 * Can now create the targets: tarfile and ftp for creating versions for
 * export.
 *
 * Fixed a number of places were tcaExitHnd was not expected to return.
 * Set the tcaSeverGlobal to 0 when the socket is closed.
 *
 * Revision 1.17  1994/04/28  16:15:25  reids
 * Changes in TCA Version 7.6:
 *  1) New functions: tcaIgnoreLogging and tcaResumeLogging
 *  2) Code for MacIntosh (MPW) version of TCA
 *
 * Revision 1.16  1994/04/26  16:23:15  rich
 * Now you can register an exit handler before anything else and it will
 * get called if connecting to central fails.
 * Also added code to handle pipe breaks during writes.
 *
 * Revision 1.15  1994/04/16  19:41:45  rich
 * First release of TCA for the DEC alpha.
 * Changes were needed because longs are 64 bits.
 * Fixed alignment assumption in the data message format.
 * Fixed the way offsets are calculated for variable length arrays.  This
 * was a problem even without 64 bit longs and pointers.
 *
 * Added the commit date to the version information printed out with the -v
 * option.
 *
 * Now uses standard defines for byte order
 * (BYTE_ORDER = BIG_ENDIAN, LITTLE_ENDIAN or PDP_ENDIAN)
 *
 * Defined alignment types: ALIGN_INT ALINE_LONGEST and ALIGN_WORD.
 *
 * *** WARNING ***
 * sending longs between alphas and non-alpha machines will probably not work.
 * *** WARNING ***
 *
 * Revision 1.14  1994/04/15  17:09:48  reids
 * Changes to support vxWorks version of TCA 7.5
 *
 * Revision 1.13  1994/04/04  16:42:48  reids
 * Fixed a bug found by Rich in waitForReply2
 *
 * Revision 1.12  1994/03/27  22:50:14  rich
 * Fixed problem with lisp version not working because of some compiler
 * flags used for the shared library version.
 * TCA now compiles for alphas, but does not run.
 *
 * Revision 1.11  1993/12/14  17:33:02  rich
 * Changed getMGlobal to GET_M_GLOBAL and changed getSGlobal to
 * GET_S_GLOBAL to conform to Chris' software standards.
 *
 * Patched problem with connecting between machines with different byte
 * orders.  The real fix requires changing the way formats are stored.
 * Searching for structural similar formats does not guarantee that you
 * find the right format.
 *
 * Revision 1.10  1993/12/01  18:57:44  rich
 * Changed TCP to  IPPROTO_TCP, a system defined constant.
 *
 * Revision 1.9  1993/12/01  18:02:53  rich
 * Fixed a problem with the port number being double converted to network
 * byte order.
 * Some general cleanup.
 *
 * Revision 1.8  1993/11/21  20:17:21  rich
 * Added shared library for sun4c_411 sunos machines.
 * Added install to the makefile.
 * Fixed problems with global variables.
 *
 * Revision 1.7  1993/10/21  16:13:36  rich
 * Fixed compiler warnings.
 *
 * Revision 1.6  1993/10/20  19:00:25  rich
 * Fixed bug with self registed messages in the lisp version.
 * Added new routine : tcaGetServerGlobal to get the server socket.
 * Fixed some bad global references for the lisp version.
 * Updated some prototypes.
 *
 * Revision 1.5  1993/08/30  21:53:12  fedor
 * V7+V6+VXWORKS Everything compiles but there are initialization problems.
 *
 * Revision 1.4  1993/08/27  07:14:19  fedor
 * First Pass at V7 and V6+VXWORKS merge
 *
 * Revision 1.3  1993/08/20  23:06:37  fedor
 * Minor changes for merge. Mostly added htons and removed cfree calls.
 *
 * Revision 1.2  1993/05/26  23:16:59  rich
 * Fixed up the comments at the top of the file.
 *
 * Revision 1.1.1.1  1993/05/20  05:45:16  rich
 * Importing tca version 8
 *
 * Revision 7.1  1993/05/20  00:29:18  rich
 * RTG - initial checkin of Chris Fedor's version 8 of tca
 *
 * Revision 1.2  1993/05/19  17:23:19  fedor
 * Added Logging.
 *
 * 27-Oct-92 Richard Goodwin, School of Computer Science, CMU
 * Changed printf to fprintf(stderr... for warning messages.
 *
 * 6-May-92 Christopher Fedor, School of Computer Science, CMU
 * Added while loop to select if select dies on an EINTR then retry select.
 *
 * 17-Feb-92 Christopher Fedor, School of Computer Science, CMU
 * Generalized call to a registered exit handler tcaExitHnd for all modules
 * except lisp. VxWorks no longer has a special tcaVxExit routine.
 *
 * 28-Oct-91 Christopher Fedor, School of Computer Science, CMU
 * Moved call for tcaVxExit so that it will be called when a close is detected
 * by both a direct connection and central disconnect. Currently the VxWorks 
 * part is not set up for direct connections and the exit call was only 
 * happening for direct connections.
 *
 * 14-Aug-91 Christopher Fedor, School of Computer Science, CMU
 * Changed VxWorks include files to vx5.0
 *
 * 25-Mar-91 Christopher Fedor, School of Computer Science, CMU
 * Added lispExitGlobal as the VxWorks way of exiting lisp.
 * Should eventually add a general abort handler and exit handler.
 *
 * 30-Jan-91 Christopher Fedor, School of Computer Science, CMU
 * Added fflush(stdout) to printf for module code calls from lisp
 *
 *  6-Dec-90 Christopher Fedor, School of Computer Science, CMU
 * Removed signal.h because the signal to call tcaClose on various signals
 * was commeted out and signal.h for sunos 4.1 eventually includes
 * sys/stdtypes which defines size_t as an int - this conflicts with the
 * vxworks includes which eventually include types.h which defines
 * size_t as a long.
 * Move sys/time.h to non VxWorks and added vxworks utime.h because the
 * timeval struct is needed and the size_t conflict exits.
 *
 *  7-Apr-90 Christopher Fedor, School of Computer Science, CMU
 * Revised to Software Standards.
 *
 *  6-Mar-89 Christopher Fedor, School of Computer Science, CMU
 * Created.  
 * fedor: 6-Mar-89 problem with many defined globals from com_server and lots
 * of repetitive code needs to be removed!  perhaps a com_com file for
 * common communications between the two - actually i want to avoid 
 * differences - module and server should be alike as much as possible.
 *
 * $Revision: 1.60 $
 * $Date: 1996/01/23 00:06:24 $
 * $Author: rich $
 *
 *****************************************************************************/

#include "globalM.h"
#include <stdarg.h>


/******************************************************************************
 *
 * FUNCTION: void printDataModuleInitialize();
 *
 * DESCRIPTION: 
 * Sets print data routines for modules.
 *
 * INPUTS: none.
 *
 * OUTPUTS: void.
 *
 *****************************************************************************/

/* ARGSUSED */
static void sampleModulePrintData(FILE *stream, const void *item, int32 next)
{
#ifdef applec
#pragma unused(stream, item, next)
#endif
  tcaModWarning( "\nWARNING: DPrint called!\n");
}

/* ARGSUSED */
static void sampleModuleMapPrintData(FILE *stream, void *map, 
				     CONST_FORMAT_PTR format,
				     int32 next, void *mapElements)
{
#ifdef applec
#pragma unused(stream, map, format, next, mapElements)
#endif
  tcaModWarning( "\nWARNING: DPrint on maps called!\n");
}

static void printDataModuleInitialize(void)
{
  GET_M_GLOBAL(dPrintSTR_FN) = (void (*)(FILE *, const char *, int32)) 
    sampleModulePrintData;
  GET_M_GLOBAL(dPrintFORMAT_FN) = (void (*)(FILE *, CONST_FORMAT_PTR, int32))
    sampleModulePrintData;
  GET_M_GLOBAL(dPrintMAP_FN) = (void (*)(FILE *, const genericMatrix *,
					 CONST_FORMAT_PTR,
					 int, const void *))
    sampleModuleMapPrintData;
  GET_M_GLOBAL(dPrintTCA_FN) = (void (*)(FILE *, const TCA_REF_TYPE *, int))
    sampleModulePrintData;
  GET_M_GLOBAL(dPrintCHAR_FN) = (void (*)(FILE *, const char *, int))
    sampleModulePrintData;
  GET_M_GLOBAL(dPrintSHORT_FN) = (void (*)(FILE *, const int16 *, int))
    sampleModulePrintData;
  GET_M_GLOBAL(dPrintLONG_FN) = (void (*)(FILE *, const long *, int))
    sampleModulePrintData;
  GET_M_GLOBAL(dPrintINT_FN) = (void (*)(FILE *, const int32 *, int))
    sampleModulePrintData;
  GET_M_GLOBAL(dPrintBOOLEAN_FN) = (void (*)(FILE *, const int32 *, int))
    sampleModulePrintData;
  GET_M_GLOBAL(dPrintFLOAT_FN) = (void (*)(FILE *, const float *, int))
    sampleModulePrintData;
  GET_M_GLOBAL(dPrintDOUBLE_FN) = (void (*)(FILE *, const double *, int))
    sampleModulePrintData;
  GET_M_GLOBAL(dPrintBYTE_FN) = (void (*)(FILE *, const char *, int))
    sampleModulePrintData;
  GET_M_GLOBAL(dPrintUBYTE_FN) = (void (*)(FILE *, const char *, int))
    sampleModulePrintData;
  GET_M_GLOBAL(dPrintTWOBYTE_FN) = (void (*)(FILE *, const char *, int))
    sampleModulePrintData;
}


/******************************************************************************
 *
 * FUNCTION: void tcaModuleInitialize()
 *
 * DESCRIPTION: 
 * Sets globals and calls main initialize routines for initializing a module.
 *
 * INPUTS: none.
 *
 * OUTPUTS: void.
 *
 *****************************************************************************/

void tcaModuleInitialize(void)
{
  if (mGlobalp() && GET_C_GLOBAL(valid)) 
    /* already have initialized the data structures */
    return;
  
  globalMInit();
  
  GET_M_GLOBAL(tcaExitHnd) = NULL;
  
  GET_M_GLOBAL(sizeDM) = 0;
  GET_M_GLOBAL(numAllocatedDM) = 0;
  
  GET_M_GLOBAL(directFlagGlobal) = FALSE;
  GET_C_GLOBAL(moduleConnectionList) = listCreate();
  
  GET_M_GLOBAL(versionGlobal).tcaMajor = TCA_VERSION_MAJOR;
  GET_M_GLOBAL(versionGlobal).tcaMinor = TCA_VERSION_MINOR;
  
  dataMsgInitialize();
  
  tcaRefInitialize();
  formatInitialize();
  
  GET_C_GLOBAL(msgIdTable) = idTableCreate("Message Table", 10);
  GET_C_GLOBAL(hndIdTable) = idTableCreate("Handler Table", 10);
  
  GET_C_GLOBAL(messageTable) = hashTableCreate(101, (HASH_FN)strHashFunc, 
					       (EQ_HASH_FN)strKeyEqFunc);
  GET_C_GLOBAL(handlerTable) = hashTableCreate(101, (HASH_FN)intStrHashFunc, 
					       (EQ_HASH_FN)intStrKeyEqFunc);
  
  GET_M_GLOBAL(classFormatTable) = hashTableCreate(11, (HASH_FN)classHashFunc, 
						   (EQ_HASH_FN) classEqFunc);
  
  printDataModuleInitialize();
  
  GET_M_GLOBAL(Message_Ignore_Set) = listCreate();
  
  GET_C_GLOBAL(valid) = TRUE;
}


/******************************************************************************
 *
 * FUNCTION: char *tcaServerMachine()
 *
 * DESCRIPTION: Simply calls getenv to read the server machine from CENTRALHOST
 *
 * INPUTS: none.
 *
 * OUTPUTS: char *
 *
 *****************************************************************************/

#ifndef VXWORKS
const char *tcaServerMachine(void)
{
  char *servHost;
  
  servHost = getenv("CENTRALHOST");
  if (!servHost) {
    tcaModWarning("Warning: CENTRALHOST environment variable not set.\n");
    return NULL;
  };
  
  return servHost;
}
#endif


/******************************************************************************
 *
 * FUNCTION: void tcaIgnoreLogging()
 *
 * DESCRIPTION: Inform central to avoid logging a particular message from 
 *              now on
 *
 * INPUTS: char *msgName to ignore.
 *
 * OUTPUTS: void.
 *
 *****************************************************************************/

void tcaIgnoreLogging(char *msgName)
{
  (void)tcaInform(TCA_IGNORE_LOGGING_INFORM, (void *)&msgName);
}

/******************************************************************************
 *
 * FUNCTION: void tcaResumeLogging()
 *
 * DESCRIPTION: Inform central to resume logging a particular message 
 *              from now on
 *
 * INPUTS: char *msgName to resume logging.
 *
 * OUTPUTS: void.
 *
 *****************************************************************************/

void tcaResumeLogging(char *msgName)
{
  (void)tcaInform(TCA_RESUME_LOGGING_INFORM, (void *)&msgName);
}

/******************************************************************************
 *
 * FUNCTION: void tcaClose()
 *
 * DESCRIPTION: 
 *
 * INPUTS: none.
 *
 * OUTPUTS: void.
 *
 *****************************************************************************/

void tcaClose(void)
{
  /* inform central that we are shutting down, if a connection exists */
  if (isValidServerConnection() &&
      (GET_C_GLOBAL(tcaServerGlobal) != CENTRAL_SERVER_ID)) {
    /*    (void)tcaInform(TCA_CLOSE_INFORM, (void *)NULL);*/
    (void)shutdown((GET_C_GLOBAL(tcaServerGlobal)), 2);
#ifdef _WINSOCK_
    closesocket((GET_C_GLOBAL(tcaServerGlobal)));
#else
    close((GET_C_GLOBAL(tcaServerGlobal)));
#endif
    FD_CLR((GET_C_GLOBAL(tcaServerGlobal)),
	   &(GET_C_GLOBAL(tcaConnectionListGlobal)));
  }
#ifndef NO_UNIX_SOCKETS
  closeSocket(GET_C_GLOBAL(listenPortNum));
#endif
  globalMInvalidate();
}

#if defined(LISP)

TCA_REF_PTR returnRootGlobal(void)
{
  return(tcaRootNode());
}

#endif


/******************************************************************************
 *
 * FUNCTION: void tcaSetContext(TCA_CONTEXT_PTR context)
 *
 * DESCRIPTION: 
 *
 * INPUTS: char *modName, *serverHost;
 *
 * OUTPUTS: void.
 *
 *****************************************************************************/
void tcaSetContext(TCA_CONTEXT_PTR context)
{
  if (context != NULL)
    GET_M_GLOBAL(currentContext) = context;
}


/******************************************************************************
 *
 * FUNCTION: TCA_CONTEXT_PTR tcaGetContext(void)
 *
 * DESCRIPTION: 
 *
 * INPUTS: 
 *
 * OUTPUTS: void.
 *
 *****************************************************************************/

TCA_CONTEXT_PTR tcaGetContext(void)
{
  return GET_M_GLOBAL(currentContext);
}


/******************************************************************************
 *
 * FUNCTION: TCA_CONTEXT_PTR tcaResetContext(void)
 *
 * DESCRIPTION: 
 *
 * INPUTS: 
 *
 * OUTPUTS: void.
 *
 *****************************************************************************/

static void tcaResetContext(void)
{
  GET_M_GLOBAL(currentContext) = 
    (TCA_CONTEXT_PTR)malloc(sizeof(TCA_CONTEXT_TYPE));
  
  FD_ZERO(&GET_C_GLOBAL(tcaConnectionListGlobal));
  FD_ZERO(&GET_C_GLOBAL(tcaListenMaskGlobal));
  GET_C_GLOBAL(servHostGlobal) = NULL;
  GET_C_GLOBAL(moduleConnectionList) = listCreate();
  GET_C_GLOBAL(msgIdTable) = idTableCreate("Message Table", 10);
  GET_C_GLOBAL(hndIdTable) = idTableCreate("Handler Table", 10);
  GET_C_GLOBAL(messageTable) = hashTableCreate(101, (HASH_FN)strHashFunc,
					       (EQ_HASH_FN)strKeyEqFunc);
  GET_C_GLOBAL(handlerTable) = hashTableCreate(101, (HASH_FN)intStrHashFunc, 
					       (EQ_HASH_FN)intStrKeyEqFunc);
  GET_C_GLOBAL(resourceTable) = hashTableCreate(11, (HASH_FN)strHashFunc, 
						(EQ_HASH_FN)strKeyEqFunc);
  GET_C_GLOBAL(tcaPendingRepliesGlobal) = listCreate();
  GET_C_GLOBAL(tcaRefFreeList) = listCreate();
  
  GET_C_GLOBAL(tcaMaxConnection) = 0;
  GET_C_GLOBAL(parentRefGlobal) = -1; 
  GET_C_GLOBAL(responseIssuedGlobal) = FALSE;
  GET_C_GLOBAL(sendMessageRefGlobal) = 1;
  GET_C_GLOBAL(listenPortNum) = 0;
  GET_C_GLOBAL(tcaListenPort) = 0;
  GET_C_GLOBAL(tcaListenSocket) = 0;
  GET_C_GLOBAL(tcaServerGlobal) = NO_SERVER_GLOBAL;
  GET_C_GLOBAL(willListen) = -1;
  GET_C_GLOBAL(valid) = FALSE;
  GET_C_GLOBAL(tappedMsgs) = NULL;
  GET_C_GLOBAL(broadcastMsgs) = NULL;
  GET_C_GLOBAL(directDefault) = FALSE;
}


/******************************************************************************
 *
 * FUNCTION: void tcaModuleProvides(resourceName, ...)
 *
 * DESCRIPTION: 
 *
 * INPUTS: 
 *
 * OUTPUTS: void.
 *
 *****************************************************************************/

void tcaModuleProvides(const char* resourceName, ...)
{
  const char *res;
  STR_LIST_PTR resources;
  va_list args;
  
  if (resourceName == NULL) 
    return;
  
  resources = strListCreate();
  res = resourceName;
  va_start(args, resourceName);
  while (res != NULL) {
    strListPushUnique(res, resources);
    res = va_arg(args,const char *);
  }
  va_end(args);
  
  tcaInform(TCA_PROVIDES_INFORM, resources);
  strListFree(&resources,FALSE);
}


/******************************************************************************
 *
 * FUNCTION: void tcaModuleProvidesArgv(resourceName, ...)
 *
 * DESCRIPTION: 
 *
 * INPUTS: 
 *
 * OUTPUTS: void.
 *
 *****************************************************************************/

void tcaModuleProvidesArgv(const char** resourceNames)
{
  int i=0;
  STR_LIST_PTR resources;
  
  if (resourceNames == NULL) 
    return;
  
  resources = strListCreate();
  for(i=0; resourceNames[i] != NULL; i++){
    strListPushUnique(resourceNames[i], resources);
  }
  
  tcaInform(TCA_PROVIDES_INFORM, resources);
  strListFree(&resources,FALSE);
}


/******************************************************************************
 *
 * FUNCTION: void tcaModuleRequires(resourceName, ...)
 *
 * DESCRIPTION: 
 *
 * INPUTS:
 *
 * OUTPUTS: void.
 *
 *****************************************************************************/

void tcaModuleRequires(const char *resourceName, ...)
{
  const char *res;
  STR_LIST_PTR resources;
  va_list args;
  
  if (resourceName == NULL) return;
  
  resources = strListCreate();
  res = resourceName;
  va_start(args, resourceName);
  while (res != NULL) {
    strListPushUnique(res, resources);
    res = va_arg(args,const char *);
  }
  va_end(args);
  
  tcaInform(TCA_REQUIRES_INFORM, resources);
  strListFree(&resources,FALSE);
}


/******************************************************************************
 *
 * FUNCTION: void tcaModuleRequiresArgv(char **resourceNames)
 *
 * DESCRIPTION: 
 *
 * INPUTS: 
 *
 * OUTPUTS: void.
 *
 *****************************************************************************/

void tcaModuleRequiresArgv(const char **resourceNames)
{
  int i;
  STR_LIST_PTR resources;
  
  if (resourceNames == NULL) return;
  
  resources = strListCreate();
  for(i=0; resourceNames[i] != NULL; i++){
    strListPushUnique(resourceNames[i], resources);
  }
  
  tcaInform(TCA_REQUIRES_INFORM, resources);
  strListFree(&resources,FALSE);
}


/******************************************************************************
 *
 * FUNCTION: const char **tcaQueryAvailable(void)
 *
 * DESCRIPTION: 
 *
 * INPUTS: 
 *
 * OUTPUTS: void.
 *
 *****************************************************************************/

char **tcaQueryAvailable(void)
{
  STR_LIST_PTR resources,tmp;
  int length,i;
  char **res;
  
  resources = strListCreate();
  tcaQuery(TCA_AVAILABLE_QUERY, NULL, resources);
  length = strListLength(resources);
  res = (char **)tcaMalloc((length + 1) * sizeof(char *));
  tmp = resources->next;
  for (i=0; i<length; i++)
    res[i] = tmp->item;
  strListFree(&resources,FALSE);
  res[length] = NULL;
  return res;
}


/******************************************************************************
 *
 * FUNCTION: void tcaConnectModule(modName, serverHost)
 *
 * DESCRIPTION: 
 *
 * INPUTS: char *modName, *serverHost;
 *
 * OUTPUTS: void.
 *
 *****************************************************************************/

void tcaConnectModule(const char *modName, const char *serverHost)
{
  MOD_DATA_TYPE modData;
  MOD_START_TYPE vData;
  DIRECT_INFO_PTR directInfo;
  TCA_RETURN_VALUE_TYPE statusV;
  char thisHost[81];
  BOOLEAN useThisHost = FALSE;
  
  /* Allow for connections to multiple central servers on different machines */
  /* Determine if already connected to another machine. */
  if (isValidServerConnection()) {
    if ((GET_C_GLOBAL(servHostGlobal) != NULL) &&
	strcmp(GET_C_GLOBAL(servHostGlobal), serverHost) == 0) {
      tcaModWarning("tcaConnectModule: already connected to %s\n", serverHost);
    } else {
      /* Connecting to another central server. */
      (GET_M_GLOBAL(modNameGlobal)) = strdup(modName);
      
      /* Need to reset the context. */
      tcaResetContext();
      if (serverHost != NULL)
	GET_C_GLOBAL(servHostGlobal) = strdup(serverHost);
      GET_C_GLOBAL(tcaServerGlobal) = NO_SERVER_GLOBAL;
      
      if (serverHost) {
	if (!connectSocket(serverHost, &(GET_C_GLOBAL(tcaServerGlobal)))) {
	  tcaModError("ERROR: Can not connect to the server on %s.",
		      serverHost);
	  return;
	}
      } else {
	GET_C_GLOBAL(tcaServerGlobal) = NO_SERVER_GLOBAL;
	bzero(&(thisHost[0]),81);
	gethostname(thisHost,80);
	useThisHost = TRUE;
	tcaModWarning("Warning: trying to connect to local host %s.\n",
		      thisHost);
	if (!connectSocket(thisHost, &(GET_C_GLOBAL(tcaServerGlobal)))) {
	  tcaModError("ERROR: Can not connect to the server on %s.",
		      thisHost);
	  GET_C_GLOBAL(tcaServerGlobal) = NO_SERVER_GLOBAL;
	  return;
	}
      }
      
      FD_SET((GET_C_GLOBAL(tcaServerGlobal)), 
	     &(GET_C_GLOBAL(tcaConnectionListGlobal)));
      
      /* Need to keep track of the maximum.  */
      GET_C_GLOBAL(tcaMaxConnection) =  MAX(GET_C_GLOBAL(tcaMaxConnection),
					    GET_C_GLOBAL(tcaServerGlobal));
      
      directInfo = NEW(DIRECT_INFO_TYPE);
      bzero(directInfo,sizeof(DIRECT_INFO_TYPE));
      directInfo->sd = (GET_C_GLOBAL(tcaServerGlobal));
      
      listInsertItem((char *)directInfo, (GET_C_GLOBAL(moduleConnectionList)));
      
      msgInfoMsgInitialize();
      classModInitialize();
      
      modData.modName = modName; /* No need to copy; not saved */
      modData.hostName = (char *)tcaMalloc(sizeof(char)*32);
#ifndef VXWORKS
      gethostname((char *)modData.hostName, 32);  
#else
      modData.hostName[0]='V';
      modData.hostName[1]='X';
      modData.hostName[2]='\0';
#endif
      
      statusV = tcaQueryCentral(TCA_CONNECT_QUERY, (void *)&modData,
				(void *)&vData);
      if (statusV != Success) {
	tcaModError("ERROR: ConnectModule Query failed.");
      }
    }
  } else {
    /* First connection. */
    tcaModuleInitialize();
    
    GET_M_GLOBAL(modNameGlobal) = strdup(modName);
    if (serverHost != NULL)
      GET_C_GLOBAL(servHostGlobal) = strdup(serverHost);
    
    FD_ZERO(&(GET_C_GLOBAL(tcaConnectionListGlobal)));
    FD_ZERO(&(GET_C_GLOBAL(tcaListenMaskGlobal)));
    GET_C_GLOBAL(tcaMaxConnection) = 0;
    
    if (serverHost) {
      if (!connectSocket(serverHost, &(GET_C_GLOBAL(tcaServerGlobal)))) {
	GET_C_GLOBAL(tcaServerGlobal) = NO_SERVER_GLOBAL;
	tcaModError("ERROR: Can not connect to the server on %s.",
		    serverHost);
	return;
      }
    } else {
      GET_C_GLOBAL(tcaServerGlobal) = NO_SERVER_GLOBAL;
      bzero(&(thisHost[0]),81);
      gethostname(thisHost,80);
      tcaModWarning("Warning: trying to connect to local host %s.\n",
		    thisHost);
      if (!connectSocket(thisHost, &(GET_C_GLOBAL(tcaServerGlobal)))) {
	tcaModError("ERROR: Can not connect to the server on %s.",
		    thisHost);
	GET_C_GLOBAL(tcaServerGlobal) = NO_SERVER_GLOBAL;
	return;
      }
    }
    
    FD_SET((GET_C_GLOBAL(tcaServerGlobal)), 
	   &(GET_C_GLOBAL(tcaConnectionListGlobal)));
    /* Need to keep track of the maximum.  */
    GET_C_GLOBAL(tcaMaxConnection) =  MAX(GET_C_GLOBAL(tcaMaxConnection),
					  GET_C_GLOBAL(tcaServerGlobal));
    
    directInfo = NEW(DIRECT_INFO_TYPE);
    bzero(directInfo,sizeof(DIRECT_INFO_TYPE));
    directInfo->sd = (GET_C_GLOBAL(tcaServerGlobal));
    
    listInsertItem((char *)directInfo, (GET_C_GLOBAL(moduleConnectionList)));
    
    msgInfoMsgInitialize();
    classModInitialize();
    
    if (!GET_M_GLOBAL(tcaRootNodeGlobal)) {
      GET_M_GLOBAL(tcaRootNodeGlobal) = CREATE_NULL_REF();
    }
    
    GET_C_GLOBAL(tcaPendingRepliesGlobal) = listCreate();
    
    modData.modName = modName; /* No need to copy; not saved */
    modData.hostName = (char *)tcaMalloc(sizeof(char)*32);
#ifndef VXWORKS
    gethostname((char *)modData.hostName, 32);  
#else
    modData.hostName[0]='V';
    modData.hostName[1]='X';
    modData.hostName[2]='\0';
#endif
    
    statusV = tcaQueryCentral(TCA_CONNECT_QUERY, (void *)&modData,
			      (void *)&vData);
    if (statusV != Success) {
      tcaModError("ERROR: ConnectModule Query failed.");
    }
    
    GET_C_GLOBAL(directDefault) = vData.direct;
    
    if (vData.version.tcaMajor != TCA_VERSION_MAJOR) {
      tcaModWarning( "*** ERROR ***\n");
      tcaModWarning( 
		    "Major Version Number is different from Task Control Server.\n");
      tcaModWarning( "Module Version     : %d.%d\n", 
		    (GET_M_GLOBAL(versionGlobal)).tcaMajor,
		    (GET_M_GLOBAL(versionGlobal)).tcaMinor);
      tcaModWarning( "Task Control Server: %d.%d\n",
		    vData.version.tcaMajor, vData.version.tcaMinor);
      tcaModError(NULL);
    } else if (vData.version.tcaMinor != TCA_VERSION_MINOR) {
      tcaModWarning( "\n*** WARNING ***\n");
      tcaModWarning( "Minor Version Number is different from Task Control Server.\n");
      tcaModWarning( "Module Version     : %d.%d\n", 
		    (GET_M_GLOBAL(versionGlobal)).tcaMajor,
		    (GET_M_GLOBAL(versionGlobal)).tcaMinor);
      tcaModWarning( "Task Control Server: %d.%d\n",
		    vData.version.tcaMajor, vData.version.tcaMinor);
      
    } else {
      tcaModWarning( "Task Control Connected %d.%d\n",
		    vData.version.tcaMajor, vData.version.tcaMinor);
    }
    
    /* RTG: Moved to tcaWaitUntilReady */
    /* modVarInitialize();*/
    
    /* dataMsgDisplayStats(); */
    
    if(GET_C_GLOBAL(directDefault)) {
      /* Make the default module direct. */
      tcaDirectResource(modName);
    }
  }
}


/******************************************************************************
 *
 * FUNCTION: void tcaWaitUntilReady()
 *
 * DESCRIPTION: 
 *
 * INPUTS: none.
 *
 * OUTPUTS: void.
 *
 *****************************************************************************/

void tcaWaitUntilReady(void)
{
  int go;
  TCA_RETURN_VALUE_TYPE statusV;
  
  if (!isValidServerConnection()) {
    tcaModError("ERROR: tcaWaitUntilReady: Must connect to server first\n");
    return;
  }
  
  /* Now initialize the tap messages. */
  modVarInitialize();
  
  tcaModWarning( "TCA Waiting ...\n");
  statusV = tcaQueryCentral(TCA_WAIT_QUERY, (void *)NULL, (void *)&go);
  if (statusV != Success) {
    tcaModError("ERROR: tcaWaitUntilReady Query failed.\n");
  }
  
  tcaModWarning( "... TCA Done.\n");
  GET_M_GLOBAL(expectedWaitGlobal) = TRUE;
}


/******************************************************************************
 *
 * FUNCTION: int tcaSingleMsgHnd(time)
 *
 * DESCRIPTION: 
 *
 * INPUTS: struct timeval *time;
 *
 * OUTPUTS: int
 *
 * NOTES:
 * Assumes that this connection is only listening to the central server
 * not very general here but it fits our current tca model of control.
 *
 * tcaConnectionListGlobal was inited from tcaServerGlobal
 *
 *****************************************************************************/

static int iterateModuleConnections(fd_set *readMask,
				    DIRECT_INFO_PTR directInfo)
{
  DATA_MSG_PTR dataMsg;
  
  if ((directInfo->sd > 0) &&(FD_ISSET(directInfo->sd, readMask))) {
    bzero((void *)&dataMsg,sizeof(dataMsg));
    switch(dataMsgRecv(directInfo->sd, &dataMsg, 0, NULL, 0)){
    case StatOK:
      execHnd(directInfo, dataMsg);
      return 1; /* 21-Apr-91: fedor: give a value to tcaHandleMessage */
    case StatError:
      tcaModError("ERROR: Reading Data Message.");
    case StatEOF:
      FD_CLR(directInfo->sd, &(GET_C_GLOBAL(tcaConnectionListGlobal)));
#ifndef VXWORKS
      (void)shutdown(directInfo->sd, 2);
#endif
#ifdef _WINSOCK_
      closesocket(directInfo->sd);
#else
      close(directInfo->sd);
#endif
      FD_CLR(directInfo->sd, &(GET_C_GLOBAL(tcaConnectionListGlobal)));
      if (!DIRECT_CONNECTION(directInfo)) {
	listDeleteItem((char *)directInfo,
		       (GET_C_GLOBAL(moduleConnectionList)));
	tcaFree((char *)directInfo);
	globalMInvalidate();
	tcaModError("Closed Connection Detected from: Central Server %s\n",
		    GET_C_GLOBAL(servHostGlobal));
      } else {
	tcaModWarning("Closed Connection Detected from: Module: sd: %d:\n", 
		      directInfo->sd);
	listDeleteItem((char *)directInfo,(GET_C_GLOBAL(moduleConnectionList)));
	tcaFree((char *)directInfo);
      }
      break;
    default:
      tcaModError("Internal Error: iterateModuleConnections: UNKNOWN INTENT\n");
    }
  }
  return TRUE;
}

/******************************************************************************
 *
 * FUNCTION: int tcaSingleMsgHnd(time)
 *
 * DESCRIPTION: 
 *
 * INPUTS: struct timeval *time;
 *
 * OUTPUTS: int
 *
 * NOTES:
 * Assumes that this connection is only listening to the central server
 * not very general here but it fits our current tca model of control.
 *
 * tcaConnectionListGlobal was inited from tcaServerGlobal
 *
 *****************************************************************************/

static int tcaSingleMsgHnd(struct timeval *time)
{
  int ret, sd;
  DIRECT_INFO_PTR directInfo;
  fd_set readMask; 
  
#ifndef VXWORKS
  if ((!GET_M_GLOBAL(directFlagGlobal)) && (time == NULL)) {
    /* Don't need to do the select, just pretend it return as it will and
     * set the return to 1.
     */
    FD_ZERO(&readMask);
    FD_SET((GET_C_GLOBAL(tcaServerGlobal)), &readMask);
    ret = 1;
  } else {
    readMask = (GET_C_GLOBAL(tcaConnectionListGlobal));
    do {
      ret = select(FD_SETSIZE, &readMask, (fd_set *)NULL,(fd_set *)NULL, time);
    }
#ifdef _WINSOCK_
    while (ret == SOCKET_ERROR && WSAGetLastError() == WSAEINTR);
#else
    while (ret < 0 && errno == EINTR);
#endif
  }
  if (ret < 1)
    return ret;
#endif
  
  if ((GET_M_GLOBAL(directFlagGlobal)) &&
      FD_ISSET((GET_C_GLOBAL(tcaListenPort)), &readMask)) {
    sd = accept((GET_C_GLOBAL(tcaListenPort)), NULL, NULL);
    if (sd < 0) {
      tcaModError("Accept Failed\n");
      return sd;
    }
    FD_SET(sd, &(GET_C_GLOBAL(tcaConnectionListGlobal)));
    /* Need to keep track of the maximum.  */
    GET_C_GLOBAL(tcaMaxConnection) =  MAX(GET_C_GLOBAL(tcaMaxConnection),
					  sd);
    
    directInfo = NEW(DIRECT_INFO_TYPE);
    bzero(directInfo,sizeof(DIRECT_INFO_TYPE));
    directInfo->sd = sd;
    
    tcaModWarning("New TCP Connection Established for a Module at sd: %d\n",
		  directInfo->sd);
    
    listInsertItem((char *)directInfo, (GET_C_GLOBAL(moduleConnectionList)));
    FD_SET(sd, &(GET_C_GLOBAL(tcaConnectionListGlobal)));
    /* Need to keep track of the maximum.  */
    GET_C_GLOBAL(tcaMaxConnection) =  MAX(GET_C_GLOBAL(tcaMaxConnection),
					  sd);
    /* 25-Jun-91: fedor: set sd in readMask as well? */
  } else if ((GET_M_GLOBAL(directFlagGlobal)) && 
	     FD_ISSET((GET_C_GLOBAL(tcaListenSocket)), &readMask)) {
    sd = accept((GET_C_GLOBAL(tcaListenSocket)), NULL, NULL);
    if (sd < 0) {
      tcaModError("Accept Failed\n");
      return sd;
    }
    FD_SET(sd, &(GET_C_GLOBAL(tcaConnectionListGlobal)));
    /* Need to keep track of the maximum.  */
    GET_C_GLOBAL(tcaMaxConnection) =  MAX(GET_C_GLOBAL(tcaMaxConnection),
					  sd);
    
    directInfo = NEW(DIRECT_INFO_TYPE);
    bzero(directInfo,sizeof(DIRECT_INFO_TYPE));
    directInfo->sd = sd;
    
    tcaModWarning("New UNIX Connection Established for a Module at sd: %d\n",
		  directInfo->sd);
    
    listInsertItem((char *)directInfo, (GET_C_GLOBAL(moduleConnectionList)));
    FD_SET(sd, &(GET_C_GLOBAL(tcaConnectionListGlobal)));
    /* Need to keep track of the maximum.  */
    GET_C_GLOBAL(tcaMaxConnection) =  MAX(GET_C_GLOBAL(tcaMaxConnection),
					  sd);
    /* 25-Jun-91: fedor: set sd in readMask as well? */
  } 
  
#if defined(VXWORKS)
  /* 26-Jun-91: fedor: VxWorks sits in a read point-to-point work
     will have to wait until this is rethought */
  FD_ZERO(&readMask);
  FD_SET((GET_C_GLOBAL(tcaServerGlobal)), &readMask);
#endif
  (void)listIterate((LIST_ITER_FN) iterateModuleConnections, &readMask, 
		    (GET_C_GLOBAL(moduleConnectionList)));
  
  return 1; /* Success */
}


/******************************************************************************
 *
 * FUNCTION: TCA_RETURN_VALUE_TYPE processQueryReply(dataMsg, msg, reply)
 *
 * DESCRIPTION: 
 * Decodes dataMsg using msg format information and storing it in reply. 
 * Reclaims dataMsg storage.
 *
 * INPUTS:
 * DATA_MSG_PTR dataMsg;
 * MSG_PTR msg;
 * char *reply;
 *
 * OUTPUTS: TCA_RETURN_VALUE_TYPE
 *
 *****************************************************************************/

static TCA_RETURN_VALUE_TYPE processQueryReply(DATA_MSG_PTR dataMsg, 
					       MSG_PTR msg,
					       void *reply, int sel)
{
  char *r2= NULL;
  int amount;
#if defined(LISP)
  BUFFER_TYPE buffer;
#endif
  
  if (!msg) {
    tcaModError( "Internal Error: processQueryReply: NULL msg.\n");
  }
  
  if (dataMsg->msgTotal == 0) {
    dataMsgFree(dataMsg);
    return NullReply;
  }
  
#if defined(LISP)
  if (reply == &(GET_M_GLOBAL(lispFlagGlobal))) {
    buffer.bstart = 0;
    buffer.buffer = NULL;
    
    if (msg->msgData->resFormat || (sel && msg->msgData->msgFormat)) {
      buffer.buffer = dataMsg->msgData;
      if (sel)
	r2 = (char *)
	  ((*(GET_M_GLOBAL(lispDecodeMsgGlobal)))(msg->msgData->msgFormat, 
						  &buffer));
      else
	r2 = (char *)
	  (*(GET_M_GLOBAL(lispDecodeMsgGlobal)))(msg->msgData->resFormat, 
						 &buffer);
    }
    else
      r2 = (char *)(*(GET_M_GLOBAL(lispDecodeMsgGlobal)))(0, NULL);      
    
    (*(GET_M_GLOBAL(lispReplyMsgGlobal)))(r2);
  } else
#endif
    {
      if (sel) {
	r2 = (char *)dataMsgDecodeMsg(msg->msgData->msgFormat, dataMsg, FALSE);
	amount = dataStructureSize(msg->msgData->msgFormat);
      }
      else {
	r2 = (char *)dataMsgDecodeMsg(msg->msgData->resFormat, dataMsg, FALSE);
	amount = dataStructureSize(msg->msgData->resFormat);
      }
      BCOPY(r2, reply, amount);
      /* Free only if it was allocated. */
      if (r2 != dataMsg->dataStruct)
	tcaFree((char *)r2);
    }
  
  dataMsgFree(dataMsg);
  return Success;
}


/******************************************************************************
 *
 * FUNCTION: TCA_RETURN_VALUE_TYPE waitForReply(ref, reply, timeout)
 *
 * DESCRIPTION: Loops to handle messages while waiting for a query reply.
 *
 * INPUTS:
 * TCA_REF_PTR ref;
 * char *reply;
 * long timeout;
 *
 * OUTPUTS: TCA_RETURN_VALUE_TYPE
 *
 * NOTES: Modified from tcaSingleMsgHnd
 *
 *****************************************************************************/

static int testMsgRef(int *refId, DATA_MSG_PTR dataMsg)
{
  if (dataMsg->msgRef == *refId)
    return TRUE;
  
  return FALSE;
}

static int iterateModuleReply(WAIT_FOR_REPLY_PTR wfr,
			      DIRECT_INFO_PTR directInfo)
{
  int sel;
  char *reply;
  TCA_REF_PTR ref;
  DATA_MSG_PTR dataMsg;
  fd_set *readMask;
  
  sel = wfr->sel;
  reply = wfr->replyData;
  ref = wfr->ref;
  readMask = wfr->readMask;
  
  if ((directInfo->sd > 0) && (FD_ISSET(directInfo->sd, readMask))) {
    FD_CLR(directInfo->sd, readMask);
    switch(dataMsgRecv(directInfo->sd, &dataMsg, ref->refId, reply, 0)){
    case StatOK:
      if (dataMsg->intent == QUERY_REPLY_INTENT) {
	if (dataMsg->msgRef == ref->refId) {
	  wfr->status = processQueryReply(dataMsg, ref->msg, reply, sel);
	  return 0;
	}
	listInsertItem((char *)dataMsg, (GET_C_GLOBAL(tcaPendingRepliesGlobal)));
      }
      else
	execHnd(directInfo, dataMsg);
      return 1;
    case StatError:
      tcaModError("ERROR: Reading Data Message in waitForReplyFrom %d.\n",
		  directInfo->sd);
    case StatEOF:
      FD_CLR(directInfo->sd, &(GET_C_GLOBAL(tcaConnectionListGlobal)));
#ifndef VXWORKS
      (void)shutdown(directInfo->sd, 2);
#ifdef _WINSOCK_
      closesocket(directInfo->sd);
#else
      close(directInfo->sd);
#endif
      FD_CLR(directInfo->sd, &(GET_C_GLOBAL(tcaConnectionListGlobal)));
      if (!DIRECT_CONNECTION(directInfo)) {
	globalMInvalidate();
	tcaModWarning(
		      "Closed Connection Detected from Central Server\n");
	listDeleteItem((char *)directInfo,
		       (GET_C_GLOBAL(moduleConnectionList)));
	tcaFree((char *)directInfo);
#if defined(LISP)
	(*(GET_M_GLOBAL(lispExitGlobal)))();
#else
	if (!(GET_M_GLOBAL(tcaExitHnd))) {
	  
#ifdef _WINSOCK_
	  WSACleanup();
#endif /* Unload Winsock DLL */
	  
	  exit(-1);
	} else {
	  (*(GET_M_GLOBAL(tcaExitHnd)))();
	}
	
#endif /* defined(LISP) */
      } else {
	tcaModWarning("Closed Connection Detected from: Module: sd: %d:\n",
		      directInfo->sd);
	listDeleteItem((char *)directInfo,
		       (GET_C_GLOBAL(moduleConnectionList)));
	tcaFree((char *)directInfo);
      }
#endif
#if defined(VXWORKS)
      (void)close(directInfo->sd);
      FD_CLR(directInfo->sd, &(GET_C_GLOBAL(tcaConnectionListGlobal)));
      if (!DIRECT_CONNECTION(directInfo)) {
	globalMInvalidate();
	tcaModWarning(
		      "Closed Connection Detected from Central Server\n");
	listDeleteItem((char *)directInfo,
		       (GET_C_GLOBAL(moduleConnectionList)));
	tcaFree((char *)directInfo);
      } else {
	tcaModWarning(
		      "Closed Connection Detected from: Module: sd: %d:\n",
		      directInfo->sd);
	listDeleteItem((char *)directInfo,
		       (GET_C_GLOBAL(moduleConnectionList)));
	tcaFree((char *)directInfo);
	
	if (!(GET_M_GLOBAL(tcaExitHnd))) {
	  
#ifdef _WINSOCK_
	  WSACleanup();
#endif /* Unload Winsock DLL */
	  
	  exit(-1);
	} else {
	  (*(GET_M_GLOBAL(tcaExitHnd)))();
	}
      }
#endif
      break;
    default:
      tcaModError("ERROR: UNKNOWN INTENT in waitForReplyFrom Response.\n");
    }
  }
  return TRUE;
}

/*****************
 * Returns Timeout if "timeout" seconds elapse without receiving a message.
 *  WAITFOREVER always waits until reply received (a very long time, at least).
 ****************/
TCA_RETURN_VALUE_TYPE waitForReplyFrom(TCA_REF_PTR ref, void *reply, 
				       int32 sel, long timeout,
				       int fd)
{
  int ret, sd;
  DATA_MSG_PTR dataMsg;
  WAIT_FOR_REPLY_TYPE wfr;
  DIRECT_INFO_PTR directInfo;
  struct timeval time;
  fd_set readMask; 
  
  if (timeout != WAITFOREVER) {
    time.tv_usec = 0;
    time.tv_sec = timeout;
  }
  
  if (!ref) {
    tcaModError("ERROR: NULL ref in waitForReplyFrom.\n");
  }
  
  for(;;) {
    dataMsg = (DATA_MSG_PTR)listMemReturnItem((LIST_ITER_FN) testMsgRef,
					      (char *)&ref->refId, 
					      (GET_C_GLOBAL(tcaPendingRepliesGlobal)));
    if (dataMsg) {
      listDeleteItem((char *)dataMsg, (GET_C_GLOBAL(tcaPendingRepliesGlobal)));
      return processQueryReply(dataMsg, ref->msg, reply, sel);
    }
    
#ifndef VXWORKS
    if (!GET_M_GLOBAL(directFlagGlobal) && (timeout == WAITFOREVER)) {
      /* Don't need to do the select, just pretend it return as it will and
       * set the return to 1.
       */
      FD_ZERO(&readMask);
      if (GET_C_GLOBAL(tcaServerGlobal) == NO_FD) {
	ret = 0;
      } else {
	FD_SET((GET_C_GLOBAL(tcaServerGlobal)), &readMask);
	ret = 1;
      }
    } else {
      if (fd == NO_FD) {
	readMask = (GET_C_GLOBAL(tcaConnectionListGlobal));
      } else {
	FD_ZERO(&readMask);
	FD_SET(fd,&readMask);
      }
      do {
	ret = select(FD_SETSIZE, &readMask, (fd_set *)NULL, (fd_set *)NULL,
		     ((timeout == WAITFOREVER) ? (struct timeval *)NULL 
		      : &time));
      }
#ifdef _WINSOCK_
      while (ret == SOCKET_ERROR && WSAGetLastError() == WSAEINTR);
#else
      while (ret < 0 && errno == EINTR);
#endif
    }
    if (ret == 0) {
      return TimeOut;
    } else if (ret < 1) {
      tcaModError("ERROR: on select in WaitForReplyFrom.\n");
    }
#endif
    
    if ((GET_M_GLOBAL(directFlagGlobal)) && 
	FD_ISSET((GET_C_GLOBAL(tcaListenPort)), &readMask)) {
      sd = accept((GET_C_GLOBAL(tcaListenPort)), 
		  (struct sockaddr *)NULL, (int *)NULL);
      if (sd < 0) {
	tcaModError("Accept Failed\n");
	return sd;
      }
      FD_SET(sd, &(GET_C_GLOBAL(tcaConnectionListGlobal)));
      /* Need to keep track of the maximum.  */
      GET_C_GLOBAL(tcaMaxConnection) =  MAX(GET_C_GLOBAL(tcaMaxConnection),
					    sd);
      
      directInfo = NEW(DIRECT_INFO_TYPE);
      bzero(directInfo,sizeof(DIRECT_INFO_TYPE));
      directInfo->sd = sd;
      
      tcaModWarning( 
		    "New Connection Established for a Module at sd: %d\n",
		    directInfo->sd);
      listInsertItem((char *)directInfo, (GET_C_GLOBAL(moduleConnectionList)));
      FD_SET(sd, &(GET_C_GLOBAL(tcaConnectionListGlobal)));
      /* Need to keep track of the maximum.  */
      GET_C_GLOBAL(tcaMaxConnection) =  MAX(GET_C_GLOBAL(tcaMaxConnection),
					    sd);
      /* 25-Jun-91: fedor: set sd in readMask as well? */
    } else if ((GET_M_GLOBAL(directFlagGlobal)) && 
	       FD_ISSET((GET_C_GLOBAL(tcaListenSocket)), &readMask)) {
      sd = accept((GET_C_GLOBAL(tcaListenSocket)), 
		  (struct sockaddr *)NULL, (int *)NULL);
      if (sd < 0) {
	tcaModError("Accept Failed\n");
	return sd;
      }
      FD_SET(sd, &(GET_C_GLOBAL(tcaConnectionListGlobal)));
      /* Need to keep track of the maximum.  */
      GET_C_GLOBAL(tcaMaxConnection) =  MAX(GET_C_GLOBAL(tcaMaxConnection),
					    sd);
      
      directInfo = NEW(DIRECT_INFO_TYPE);
      bzero(directInfo,sizeof(DIRECT_INFO_TYPE));
      directInfo->sd = sd;
      
      tcaModWarning( 
		    "New Connection Established for a Module at sd: %d\n",
		    directInfo->sd);
      listInsertItem((char *)directInfo, (GET_C_GLOBAL(moduleConnectionList)));
      FD_SET(sd, &(GET_C_GLOBAL(tcaConnectionListGlobal)));
      /* Need to keep track of the maximum.  */
      GET_C_GLOBAL(tcaMaxConnection) =  MAX(GET_C_GLOBAL(tcaMaxConnection),
					    sd);
      /* 25-Jun-91: fedor: set sd in readMask as well? */
    }
    
#if defined(VXWORKS)
    /* 26-Jun-91: fedor: VxWorks sits in a read point-to-point work
       will have to wait until this is rethought */
    FD_ZERO(&readMask);
    FD_SET((GET_C_GLOBAL(tcaServerGlobal)), &readMask);
#endif
    
    wfr.sel = sel;
    wfr.readMask = &readMask;
    wfr.ref = ref;
    wfr.replyData = reply;
    
    if (!listIterate((LIST_ITER_FN) iterateModuleReply, (char *)&wfr, 
		     (GET_C_GLOBAL(moduleConnectionList)))) {
      return wfr.status;
    }
  }
}

TCA_RETURN_VALUE_TYPE waitForReply(TCA_REF_PTR ref, void *reply)
{
  return waitForReplyFrom(ref, reply, 0, WAITFOREVER, NO_FD);
}


/******************************************************************************
 *
 * FUNCTION: TCA_RETURN_VALUE_TYPE tcaHandleMessage(sec)
 *
 * DESCRIPTION: 
 *
 * INPUTS: long sec;
 *
 * OUTPUTS: TCA_RETURN_VALUE_TYPE
 *
 *****************************************************************************/

TCA_RETURN_VALUE_TYPE tcaHandleMessage(long sec)
{
  int ret;
#if !defined(VXWORKS)
  struct timeval time;
#endif
  
  if (!isValidServerConnection()) {
    tcaModError("ERROR: Must connect to server first.\n");
    return Failure;
  } else if (!GET_M_GLOBAL(expectedWaitGlobal)) {
    tcaModWarning(
		  "WARNING: call tcaWaitUntilReady before listening for messages\n");
  }
  
#if defined(VXWORKS)
  ret = tcaSingleMsgHnd(NULL);
#else
  
  if (sec == WAITFOREVER)
    ret = tcaSingleMsgHnd((struct timeval *)NULL);
  else {
    time.tv_sec = sec;
    time.tv_usec = 0;
    ret = tcaSingleMsgHnd(&time);
  }
#endif
  
  switch(ret) {
  case 0:
    return TimeOut;
  case -1:
    return Failure;
  default:
    return Success;
  }
}


/******************************************************************************
 *
 * FUNCTION: void tcaModuleListen()
 *
 * DESCRIPTION: 
 *
 * INPUTS: none.
 * 
 * OUTPUTS: void.
 *
 *****************************************************************************/

void tcaModuleListen(void)
{
  
  if (!isValidServerConnection()) {
    tcaModError("ERROR: Must connect to server first (tcaModuleListen).\n");
    return;
  } else if (!GET_M_GLOBAL(expectedWaitGlobal)) {
    tcaModWarning(
		  "WARNING: call tcaWaitUntilReady before listening for messages\n");
  }
  
  for(;;) {
    tcaSingleMsgHnd((struct timeval *)NULL);
  }
}


/******************************************************************************
 *
 * FUNCTION: void tcaRegisterExitHandler(proc)
 *
 * DESCRIPTION: 
 *
 * INPUTS:
 * int (*proc)();
 * 
 * OUTPUTS: void.
 *
 *****************************************************************************/

void tcaRegisterExitProc(void (*proc)(void))
{
  /* make sure the data structures are initialzied. */
  tcaModuleInitialize();
  (GET_M_GLOBAL(tcaExitHnd)) = proc;
}

/****************************************************************************/

void tcaDirectResource(const char *name)
{
  int port;
  DIRECT_TYPE direct;
  
  if (GET_C_GLOBAL(tcaListenPort) ==0) {
    port = LISTEN_PORT;
    while ((port < LISTEN_PORT_MAX) && 
	   !listenAtPort(&port, &(GET_C_GLOBAL(tcaListenPort))))
      port++;
    
    if (port == LISTEN_PORT_MAX) {
      tcaModError( "ERROR: Exceeded available ports for direct connection.\n");
    }
    
#ifndef NO_UNIX_SOCKETS
    listenAtSocket(port, &(GET_C_GLOBAL(tcaListenSocket)));
#endif
    
    (GET_M_GLOBAL(directFlagGlobal)) = TRUE;
    FD_SET((GET_C_GLOBAL(tcaListenPort)), 
	   &(GET_C_GLOBAL(tcaConnectionListGlobal)));
    /* Need to keep track of the maximum.  */
    GET_C_GLOBAL(tcaMaxConnection) =  MAX(GET_C_GLOBAL(tcaMaxConnection),
					  (GET_C_GLOBAL(tcaListenPort)));
    FD_SET((GET_C_GLOBAL(tcaListenSocket)), 
	   &(GET_C_GLOBAL(tcaConnectionListGlobal)));
    /* Need to keep track of the maximum.  */
    GET_C_GLOBAL(tcaMaxConnection) =  MAX(GET_C_GLOBAL(tcaMaxConnection),
					  (GET_C_GLOBAL(tcaListenSocket)));
    GET_C_GLOBAL(listenPortNum) = port;
  }
  
  direct.port = GET_C_GLOBAL(listenPortNum);
  direct.name = name;
  
  tcaInform(TCA_DIRECT_RES_INFORM, &direct);
}

/****************************************************************************/

static int testInfo(DIRECT_INFO_PTR directInfo, DIRECT_INFO_PTR infoItem)
{
  if (directInfo->port == infoItem->port &&
      strKeyEqFunc(directInfo->host, infoItem->host))
    return TRUE;
  else
    return FALSE;
}

void establishDirect(MSG_PTR msg)
{
  int sd;
  DIRECT_INFO_PTR directInfo, oldInfo;
  
  directInfo = NEW(DIRECT_INFO_TYPE);
  bzero(directInfo,sizeof(DIRECT_INFO_TYPE));
  
  (void)tcaQueryCentral(TCA_DIRECT_INFO_QUERY, (void *)&msg->msgData->name,
			(void *)directInfo);
  
  if (strcmp(GET_M_GLOBAL(modNameGlobal),directInfo->module) == 0) {
    tcaFreeReply(TCA_DIRECT_INFO_QUERY, (void *)directInfo);
    tcaModWarning( "Using central connection for %s to %s.\n",
		  msg->msgData->name, GET_M_GLOBAL(modNameGlobal));
    return;
  } else {
    msg->direct = TRUE;
    msg->intent = directInfo->intent;
    
    oldInfo = (DIRECT_INFO_PTR)
      listMemReturnItem((LIST_ITER_FN) testInfo,
			(char *)directInfo,
			(GET_C_GLOBAL(moduleConnectionList)));
    
    if (oldInfo) {
      msg->sd = oldInfo->sd;
      tcaFreeReply(TCA_DIRECT_INFO_QUERY, (void *)directInfo);
      tcaModWarning( "Using existing direct connection for %s to %s.\n",
		    msg->msgData->name, oldInfo->module);
      return;
    } 
    
    if (!connectAt(directInfo->host, directInfo->port, &sd)) {
      tcaModError("ERROR: Can not establish direct connection to %s at %ld.\n",
		  directInfo->host, (long)directInfo->port);
    } else {
      (GET_M_GLOBAL(directFlagGlobal)) = TRUE;
      tcaModWarning( "Direct Connection Established for %s to %s.\n", 
		    msg->msgData->name, directInfo->module);
    }
    FD_SET(sd, &(GET_C_GLOBAL(tcaConnectionListGlobal)));  
    /* Need to keep track of the maximum.  */
    GET_C_GLOBAL(tcaMaxConnection) =  MAX(GET_C_GLOBAL(tcaMaxConnection),
					  sd);
  }
  
  directInfo->sd = sd;
  
  /* 26-Jun-91: fedor
   * attempt to make the server the first item
   * otherwise things get overrun on the server socket
   * in waitForReply - going to have to redesign this nonsense!
   */
  listInsertItemLast(directInfo, (GET_C_GLOBAL(moduleConnectionList)));
  
  msg->sd = sd;
}


/******************************************************************************
 *
 * FUNCTION: int tcaGetVersionMajor(void)
 *
 * DESCRIPTION: 
 *
 * INPUTS:
 * 
 * OUTPUTS: version number.
 *
 *****************************************************************************/

int32 tcaGetVersionMajor(void)
{
  return TCA_VERSION_MAJOR;
}


/******************************************************************************
 *
 * FUNCTION: int tcaGetVersionMinor(void)
 *
 * DESCRIPTION: 
 *
 * INPUTS:
 * 
 * OUTPUTS: version number.
 *
 *****************************************************************************/

int32 tcaGetVersionMinor(void)
{
  return TCA_VERSION_MINOR;
}
