/*****************************************************************************
 * PROJECT: TCA
 *
 * (c) Copyright 1996 Reid Simmons. All rights reserved.
 *
 * FILE: script.c
 *
 * ABSTRACT: Run a script file.
 *
 * $Source: /afs/cs.cmu.edu/project/TCA/Master/tcaV8/tools/nanny/script.c,v $
 * $Revision: 1.1 $
 * $Date: 1996/01/22 21:31:27 $
 * $Author: reids $
 *
 * REVISION HISTORY:
 *
 * $Log: script.c,v $
 * Revision 1.1  1996/01/22  21:31:27  reids
 * Fixed the way stdout is handled, using pseudo-terminals, so that it preserves
 *   the line-buffering mode of the real TTY.
 * Added support for running scripts (not really integrated, yet, but initial
 *   tests are working).
 *
 *
 ****************************************************************/

#include "tca/libc.h"
#include "tca/basics.h"
#include "tca/timeUtils.h"
#include "tca/devUtils.h"

#include "nannyCommon.h"
#include "nannyDev.h"
#include "messages.h"
#include "runConsole.h"
#include "script.h"

static SCRIPT_LIST_PTR activeListG = NULL;

/* Forward References */
static void processConditional (SCRIPT_STATEMENT_PTR statement);

static time_t currentSeconds (void)
{
  struct timeval time;

  gettimeofday(&time, NULL);
  return time.tv_sec;
}

void setLocalTimeFromAbsTime(struct tm *localTime, ABS_TIME_PTR absTime)
{
  if (TIME_IS_SET(absTime->mth)) localTime->tm_mon = absTime->mth;
  if (TIME_IS_SET(absTime->day)) localTime->tm_mday = absTime->day;
  if (TIME_IS_SET(absTime->hr))  localTime->tm_hour = absTime->hr;
  if (TIME_IS_SET(absTime->min)) localTime->tm_min = absTime->min;
  if (TIME_IS_SET(absTime->sec)) localTime->tm_sec = absTime->sec;
}

static time_t nextTime(time_t currentTime, SCRIPT_TIME_PTR scriptTime)
{
  struct tm *localTime;
  time_t nextTime;

  if (scriptTime->useRelative) {
    return scriptTime->relative + currentTime;
  } else {
    localTime = localtime(&currentTime);
    setLocalTimeFromAbsTime(localTime, &scriptTime->absolute);
    nextTime = timelocal(localTime);
    if (nextTime <= currentTime) {
      /* Increment time to the *next* matching absolute time (e.g., next
	 year if the month is set, next month if the day is set, etc.) */
      if (TIME_IS_SET(scriptTime->absolute.mth)) {
	localTime->tm_year++;
	nextTime = timelocal(localTime);
      } else if (TIME_IS_SET(scriptTime->absolute.day)) {
	localTime->tm_mon++;
	if (localTime->tm_mon > 11) {
	  localTime->tm_mon = 0;
	  localTime->tm_year++;
	}
	nextTime = timelocal(localTime);
      } else if (TIME_IS_SET(scriptTime->absolute.day)) {
	nextTime += 24*60*60;
      } else if (TIME_IS_SET(scriptTime->absolute.min)) {
	nextTime += 60*60;
      } else if (TIME_IS_SET(scriptTime->absolute.sec)) {
	nextTime += 60;
      }
    }
    return nextTime;
  }
}

static void setAlarmTime(SCRIPT_TIME_PTR scriptTime, TIME_ITER_PTR timeData)
{
  scriptTime->alarm = nextTime(timeData->current, scriptTime);
  if (timeData->alarm > scriptTime->alarm) {
    timeData->alarm = scriptTime->alarm;
  }
}

static struct timeval currentAlarmTime = {TIME_MAX, 0};

static void clearAlarm (void)
{
  currentAlarmTime.tv_sec = TIME_MAX;
  devSetAlarm(theLocalDevice()->dev, &currentAlarmTime, (Handler)NULL, NULL);
}

static void resetAlarm (time_t alarmTime)
{
  if (alarmTime < currentAlarmTime.tv_sec) {
    currentAlarmTime.tv_sec = alarmTime;
    devSetAlarm(theLocalDevice()->dev, &currentAlarmTime,
		(Handler)processAtConditionals, NULL);
  }
}

static void initTimeData (TIME_ITER_PTR timeData)
{
  timeData->current = currentSeconds();
  timeData->alarm = TIME_MAX;
}

static BOOLEAN activateStatement(TIME_ITER_PTR timeData,
				 SCRIPT_STATEMENT_PTR statement)
{
  static char buf[DEFAULT_LINE_LENGTH];

  switch (statement->type) {
  case SendAction: 
    sprintf(buf, "%s\n", statement->line);
    runCommandProcess(statement->process, buf);
    if (convertProcessNameToId(statement->process) == currentProc) {
      EditAppend (buf, strlen(buf));
      WFlush();
    }
    break;

  case RunAction:      runNewProcess(statement->process); break;
  case KillAction:     runKillProcess(statement->process); break;
  case SuspendAction:  runSuspendProcess(statement->process); break;
  case ContinueAction: runContinueProcess(statement->process); break;

  case WhenReadyConditional: 
    /* If process is in process table, it has been announced -- is that
       always the same as being ready? */
    if (convertProcessNameToId(statement->process) != -1) {
      processConditional(statement);
    } else {
      ACTIVATE(statement);
    }
    break;

  case WhenConditional: 
    ACTIVATE(statement);
    /* Make sure loudnes of process is VERBOSE */
    messageMakeVerbose(buf, statement->process);
    (void)safePassLocalMessage(buf);
    break;

  case AtConditional:
    ACTIVATE(statement);
    setAlarmTime(&statement->time, timeData);
    break;

  default: WARNING2("Uknown statement type %d\n", statement->type);
  }
  return TRUE;
}

static void activateStatements (SCRIPT_LIST_PTR script)
{
  TIME_ITER_TYPE timeData;

  initTimeData(&timeData);
  ITERATE_SCRIPT_LIST(activateStatement, &timeData, script);
  resetAlarm(timeData.alarm);
}

static void processConditional (SCRIPT_STATEMENT_PTR statement)
{
  if (!statement->continual) DEACTIVATE(statement);
  if (statement->subStatements) activateStatements(statement->subStatements);
}

static BOOLEAN atHandler (TIME_ITER_PTR timeData, 
			  SCRIPT_STATEMENT_PTR statement)
{
  if (statement->type == AtConditional && 
      statement->time.alarm <= timeData->current) {
    clearAlarm();
    processConditional(statement);
    setAlarmTime(&statement->time, timeData);
  }
  return TRUE;
}

void processAtConditionals (void)
{
  TIME_ITER_TYPE timeData;

  initTimeData(&timeData);
  ITERATE_ACTIVE_LIST(atHandler, &timeData);
  resetAlarm(timeData.alarm);
}

static BOOLEAN whenHandler (WHEN_ITER_PTR whenData, 
			    SCRIPT_STATEMENT_PTR statement)
{
  if (statement->type == WhenConditional && 
      !strcmp(statement->process, whenData->process) &&
      strstr(whenData->line, statement->line)) {
    processConditional(statement);
  }
  return TRUE;
}

void processWhenConditionals (char *process, char *line)
{
  WHEN_ITER_TYPE whenData;

  whenData.process = process;   whenData.line = line; 
  ITERATE_ACTIVE_LIST(whenHandler, &whenData);
}

static BOOLEAN whenReadyHandler (WHEN_ITER_PTR whenData, 
				 SCRIPT_STATEMENT_PTR statement)
{
  if (statement->type == WhenReadyConditional && 
      !strcmp(statement->process, whenData->process)) {
    processConditional(statement);
  }
  return TRUE;
}

void processWhenReadyConditionals (char *process)
{
  WHEN_ITER_TYPE whenData;

  whenData.process = process;   whenData.line = NULL;
  ITERATE_ACTIVE_LIST(whenReadyHandler, &whenData);
}

void runScript (SCRIPT_LIST_PTR script)
{
  if (!activeListG) activeListG = listCreate();
  activateStatements(script);
}
