/*****************************************************************************
 * PROJECT: TCA
 *
 * (c) Copyright 1996 Reid Simmons. All rights reserved.
 *
 * FILE: scriptParse.c
 *
 * ABSTRACT: Parse a script file and create the appropriate data structures.
 *
 * $Source: /afs/cs.cmu.edu/project/TCA/Master/tcaV8/tools/nanny/scriptParse.c,v $
 * $Revision: 1.1 $
 * $Date: 1996/01/22 21:31:32 $
 * $Author: reids $
 *
 * REVISION HISTORY:
 *
 * $Log: scriptParse.c,v $
 * Revision 1.1  1996/01/22  21:31:32  reids
 * Fixed the way stdout is handled, using pseudo-terminals, so that it preserves
 *   the line-buffering mode of the real TTY.
 * Added support for running scripts (not really integrated, yet, but initial
 *   tests are working).
 *
 *
 ****************************************************************/

#include "tca/libc.h"
#include "tca/basics.h"
#include "tca/timeUtils.h"

#include "nannyCommon.h"
#include "script.h"
#include "scriptParse.h"

/* Forward References */
static void printScript (FILE *file, SCRIPT_LIST_PTR script, int indent);

static void scriptInsertStatement (SCRIPT_STATEMENT_PTR statement,
				   SCRIPT_LIST_PTR *listHandle)
{
  if (*listHandle == NULL) *listHandle = listCreate();
  ADD_STATEMENT(statement, *listHandle);
}

static SCRIPT_STATEMENT_PTR scriptCreateStatement (SCRIPT_STATEMENT_ENUM type)
{
  SCRIPT_STATEMENT_PTR statement;

  statement = (SCRIPT_STATEMENT_PTR)malloc(sizeof(SCRIPT_STATEMENT_TYPE));
  statement->type = type;
  statement->continual = FALSE;
  statement->process = statement->line = "";
  statement->time.useRelative = TRUE;
  statement->time.relative = statement->time.alarm = TIME_MAX;
  statement->time.absolute.mth = -1;
  statement->time.absolute.day = statement->time.absolute.hr = -1;
  statement->time.absolute.min = statement->time.absolute.sec = -1;
  statement->subStatements = NULL;

  return statement;
}

static SCRIPT_STATEMENT_PTR scriptCreateAction (SCRIPT_STATEMENT_ENUM type,
						char *process)
{
  SCRIPT_STATEMENT_PTR statement;

  statement = scriptCreateStatement(type);
  statement->process = process;

  return statement;
}

static SCRIPT_STATEMENT_PTR scriptCreateSend (char *process, char *line)
{
  SCRIPT_STATEMENT_PTR statement;

  statement = scriptCreateStatement(SendAction);
  statement->process = process;
  statement->line = line;

  return statement;
}

static SCRIPT_STATEMENT_PTR scriptCreateWhenReady (char *process)
{
  SCRIPT_STATEMENT_PTR statement;

  statement = scriptCreateStatement(WhenReadyConditional);
  statement->process = process;

  return statement;
}

static SCRIPT_STATEMENT_PTR scriptCreateWhen (char *process, BOOLEAN continual,
					      char* line)
{
  SCRIPT_STATEMENT_PTR statement;

  statement = scriptCreateStatement(WhenConditional);
  statement->process = process;
  statement->line = line;
  statement->continual = continual;

  return statement;
}

static SCRIPT_STATEMENT_PTR scriptCreateRelAt (BOOLEAN continual, time_t delay)
{
  SCRIPT_STATEMENT_PTR statement;

  statement = scriptCreateStatement(AtConditional);
  statement->continual = continual;
  statement->time.useRelative = TRUE;
  statement->time.relative = delay;

  return statement;
}

static SCRIPT_STATEMENT_PTR scriptCreateAbsAt (BOOLEAN continual, 
					       short mth, short day,
					       short hr, short min, short sec)
{
  SCRIPT_STATEMENT_PTR statement;

  statement = scriptCreateStatement(AtConditional);
  statement->continual = continual;
  statement->time.useRelative = FALSE;
  statement->time.absolute.mth = mth;
  statement->time.absolute.day = day;
  statement->time.absolute.hr = hr;
  statement->time.absolute.min = min;
  statement->time.absolute.sec = sec;

  return statement;
}

SCRIPT_LIST_PTR scriptParse (FILE *scriptFile)
{
  return NULL;
}

static void printScriptTime (FILE *file, SCRIPT_TIME_PTR time, int indent)
{
  char buf[DEFAULT_LINE_LENGTH];
  struct tm *localTime;
  time_t zeroTime = 0;

  if (time->useRelative) {
    localTime = gmtime(&time->relative);
    strftime(buf, DEFAULT_LINE_LENGTH, "+%k:%M:%S", localTime);
  } else {
    localTime = gmtime(&zeroTime);
    setLocalTimeFromAbsTime(localTime, &time->absolute);
    if (TIME_IS_SET(time->absolute.mth) || TIME_IS_SET(time->absolute.day)) {
      strftime(buf, DEFAULT_LINE_LENGTH, "%h-%d %k:%M:%S", localTime);
    } else {
      strftime(buf, DEFAULT_LINE_LENGTH, "%k:%M:%S", localTime);
    }
  }
  fprintf(file, buf);
}

static void printIndent (FILE *file, int indent)
{
  int i;

  for (i=0; i<indent; i++) fprintf(file, " ");
}

static void printSubElements (FILE *file, SCRIPT_STATEMENT_PTR statement,
			      int indent)
{
  int num;

  num = listLength(statement->subStatements);
  if (num > 0) {
    if (num > 1) { printIndent(file, indent); fprintf(file, "BEGIN:\n"); }
    printScript(file, statement->subStatements,  indent+INDENT_INCREMENT);
    if (num > 1) { printIndent(file, indent); fprintf(file, "END:\n"); }
  }
}

static void printScript (FILE *file, SCRIPT_LIST_PTR script, int indent)
{
  SCRIPT_STATEMENT_PTR statement;

  statement = (SCRIPT_STATEMENT_PTR)listFirst(script);
  while (statement) {
    printIndent(file, indent);
    switch (statement->type) {
    case SendAction: 
      fprintf(file, "SEND: %s \"%s\"\n", statement->process, statement->line);
      break;
    case RunAction:
      fprintf(file, "RUN: %s\n", statement->process); break;
    case KillAction:
      fprintf(file, "KILL: %s\n", statement->process); break;
    case SuspendAction:
      fprintf(file, "SUSPEND: %s\n", statement->process); break;
    case ContinueAction:
      fprintf(file, "CONTINUE: %s\n", statement->process); break;
    case WhenReadyConditional:
      fprintf(file, "WHENREADY: %s\n", statement->process); 
      printSubElements(file, statement, indent);
      break;
    case WhenConditional:
      fprintf(file, "WHEN%s: %s \"%s\"\n",
	      (statement->continual ? "EVERY" : ""),
	      statement->process, statement->line);
      printSubElements(file, statement, indent);
      break;
    case AtConditional:
      fprintf(file, "AT%s: ", (statement->continual ? "EVERY" : ""));
      printScriptTime(file, &statement->time, indent);
      fprintf(file, "\n");
      printSubElements(file, statement, indent);
      break;
    default:;
    }
    statement = (SCRIPT_STATEMENT_PTR)listNext(script);
  }
}

void scriptPrint (SCRIPT_LIST_PTR script)
{
  printScript(stdout, script, 0);
}

#define TEST
#ifdef TEST
/*
 1. Start up a gdb process:
 RUN: gCTR
 WHENREADY: gCTR
 SEND: gCTR r -x -2
*/
SCRIPT_LIST_PTR testScriptA (void)
{
  SCRIPT_LIST_PTR script = NULL;
  SCRIPT_STATEMENT_PTR statement, statement2;

  statement = scriptCreateAction(RunAction, "gCTR");
  scriptInsertStatement(statement, &script);

  statement = scriptCreateWhenReady("gCTR");
  scriptInsertStatement(statement, &script);

  statement2 = scriptCreateSend("gCTR", "r -x -2 -s jubilee");
  scriptInsertStatement(statement2, &statement->subStatements);

  return script;
}

/*
2. Say "hi" to Greg every 10 seconds:
 RUN: CTS
 WHENREADY: CTS
 ATEVERY: +:10
 SEND: CTS say Hi Greg
*/
SCRIPT_LIST_PTR testScriptB (void)
{
  SCRIPT_LIST_PTR script = NULL;
  SCRIPT_STATEMENT_PTR statement, statement2, statement3;

  statement = scriptCreateAction(RunAction, "CTS");
  scriptInsertStatement(statement, &script);

  statement = scriptCreateWhenReady("CTS");
  scriptInsertStatement(statement, &script);

  statement2 = scriptCreateRelAt(TRUE, 10);
  scriptInsertStatement(statement2, &statement->subStatements);

  statement3 = scriptCreateSend("CTS", "say Hi Greg");
  scriptInsertStatement(statement3, &statement2->subStatements);

  return script;
}

/*
3. Wait until noon, then say "hi" to Greg every 30 seconds:
 AT: 12:00:00
 BEGIN:
  RUN: CTS
  WHENREADY: CTS
  ATEVERY: +:30
  SEND: CTS say Hi Greg
 END:
*/
SCRIPT_LIST_PTR testScriptC (void)
{
  SCRIPT_LIST_PTR script = NULL;
  SCRIPT_STATEMENT_PTR statement;

  statement = scriptCreateAbsAt(FALSE, -1, -1, 12, 0, 0);
  scriptInsertStatement(statement, &script);

  statement->subStatements = testScriptB();

  return script;
}

/*
4. Tell central to resend messages, and kill CTR after three bumps.
 RUN: central
 RUN: CTR
# Note: Don't have to put a "whenready" central, because CTR depends on central
 WHENREADY: CTR
 BEGIN:
  WHENEVERY: central Resend messages?
  SEND: central Y

  WHEN: CTR Bump Recovery
  WHEN: CTR Bump Recovery
  WHEN: CTR Bump Recovery
  KILL: CTR
 END:
*/
SCRIPT_LIST_PTR testScriptD (void)
{
  SCRIPT_LIST_PTR script = NULL;
  SCRIPT_STATEMENT_PTR statement, statement2, statement3, statement4, statement5;

  statement = scriptCreateAction(RunAction, "central");
  scriptInsertStatement(statement, &script);

  statement = scriptCreateAction(RunAction, "CTR");
  scriptInsertStatement(statement, &script);

  statement = scriptCreateWhenReady("CTR");
  scriptInsertStatement(statement, &script);

  statement2 = scriptCreateWhen("central", TRUE, "Resend messages?");
  scriptInsertStatement(statement2, &statement->subStatements);

  statement3 = scriptCreateSend("central", "Y");
  scriptInsertStatement(statement3, &statement2->subStatements);

  statement2 = scriptCreateWhen("CTR", FALSE, "BUMP RECOVERY");
  scriptInsertStatement(statement2, &statement->subStatements);

  statement3 = scriptCreateWhen("CTR", FALSE, "BUMP RECOVERY");
  scriptInsertStatement(statement3, &statement2->subStatements);

  statement4 = scriptCreateWhen("CTR", FALSE, "BUMP RECOVERY");
  scriptInsertStatement(statement4, &statement3->subStatements);

  statement5 = scriptCreateAction(KillAction, "CTR");
  scriptInsertStatement(statement5, &statement4->subStatements);

  return script;
}

#endif
